#include <cmath>

#include "common/utility.h"
#include "common/image.h"
#include <cmath>
using namespace ipol;
 
ImageF::ImageF():m_buf(0),m_size(0,0,0)
{
}

ImageF::ImageF(const ImageSize& size):m_buf(size.m_width*size.m_height*size.m_nc), m_size(size)
{
}

ImageF::ImageF(float* buf, const ImageSize& size, bool interleaved):m_buf(size.m_width*size.m_height*size.m_nc), m_size(size)
{
    if(!interleaved)
    {
        int nb=size.m_width*size.m_height*size.m_nc;
        for (int i = 0; i < nb; ++i)
        {
            m_buf.get()[i]=buf[i];
        }
    }
    else
    {
        for (int channel=0; channel<size.m_nc; channel++)
        {
            float* ch=m_buf.get()+channel*size.m_width*size.m_height;
            // Copy the current channel
            for (int pix=0; pix<size.m_width*size.m_height; pix++)
            ch[pix] = buf[size.m_nc*pix+channel];
        }
    }
}

ImageF::ImageF(SmartArray<float> buf, const ImageSize& size):m_buf(buf), m_size(size)
{
}

ImageF::~ImageF()
{
}

ImageF::ImageF(const ImageF& other):m_buf(other.m_buf),m_size(other.m_size)
{
}


ImageF&
ImageF::operator=(const ImageF& other)
{
    if(this!=&other)
    {
        m_buf=other.m_buf;
        m_size=other.m_size;
    }
    return *this;
}

ImageF
ImageF::clone() const
{
    return ImageF(m_buf.get(),m_size);
}

const ImageSize&
ImageF::getSize() const
{
    return m_size;
}

int
ImageF::getNumPixels() const
{
    return m_size.m_width*m_size.m_height;
}

float*
ImageF::getData() const
{
    return m_buf.get();
}

SmartArray<float>
ImageF::getInterleavedData() const
{
    int w= m_size.m_width, h=m_size.m_height, nc= m_size.m_nc;
    SmartArray<float> result(w*h*nc);
    interleave(m_buf.get(), result.get(), w, h, nc);
    return result;
}

float&
ImageF::operator [](int k)
{
    return m_buf.get()[k];
}

float
ImageF::operator [](int k) const
{
    return m_buf.get()[k];
}


float&
ImageF::operator()(int x, int y, int c)
{
    return (*this)[x+m_size.m_width*y+m_size.m_width*m_size.m_height*c];
}

float
ImageF::operator()(int x, int y, int c) const
{
    return (*this)[x+m_size.m_width*y+m_size.m_width*m_size.m_height*c];
}

ImageF
ImageF::operator+(const ImageF& other) const
{
    const ImageSize& size=this->getSize();
    ImageF result(size);
    for (int pix=0; pix<size.m_width*size.m_height*size.m_nc; pix++)
            result[pix] = m_buf[pix]+other.m_buf[pix];
    return result;
}

ImageF
ImageF::operator-(const ImageF& other) const
{
    const ImageSize& size=this->getSize();
    ImageF result(size);
    for (int pix=0; pix<size.m_width*size.m_height*size.m_nc; pix++)
            result[pix] = m_buf[pix]-other.m_buf[pix];
    return result;
}

ImageF
ImageF::operator*(const ImageF& other) const
{
    const ImageSize& size=this->getSize();
    ImageF result(size);
    for (int pix=0; pix<size.m_width*size.m_height*size.m_nc; pix++)
            result[pix] = m_buf[pix]*other.m_buf[pix];
    return result;
}

ImageF
ImageF::transpose() const
{
    ImageSize outSize=getSize();
    std::swap(outSize.m_width, outSize.m_height);
    ImageF output(outSize);
    for (int x = 0; x < outSize.m_width; ++x)
        for (int y = 0; y < outSize.m_height; ++y)
            for (int c = 0; c < outSize.m_nc; ++c)
            {
                output(x,y,c)=(*this)(y,x,c);
            }
    return output;
}

ImageF
ImageF::crop(int x1, int x2, int y1, int y2, int c1, int c2) const
{
    x2=(x2<0)?m_size.m_width:x2;
    y2=(y2<0)?m_size.m_height:y2;
    c2=(c2<0)?m_size.m_nc:c2;
    ImageSize outSize(x2-x1, y2-y1, c2-c1);
    ImageF output(outSize);
    for (int x = x1; x < x2; ++x)
        for (int y = y1; y < y2; ++y)
            for (int c = c1; c < c2; ++c)
            {
                output(x-x1,y-y1,c-c1)=(*this)(x,y,c);
            }
    return output;
}

ImageF
ImageF::centercrop(int width, int height, int c1, int c2) const
{
    int x1 = (m_size.m_width - width)/2,
        y1 = (m_size.m_height - height)/2,
        x2 = x1 + width,
        y2 = y1 + height;
    return crop(x1, x2, y1, y2, c1, c2);
}

ImageF
ImageF::pow(float exp) const
{
    const ImageSize& size=this->getSize();
    ImageF result(size);
    for (int pix=0; pix<size.m_width*size.m_height*size.m_nc; pix++)
        result[pix] = std::pow(m_buf[pix],exp);
    return result;
}

