#ifndef _IMAGE_H_
#define _IMAGE_H_

#include <memory>
#include <string>
#include <cassert>

namespace ipol
{
    template<typename T>
    struct my_array_deleter
    {
        void operator()(T* p)
        {
            delete [] p;
        }
    };
    
    template<typename T>
    class SmartArray: public std::shared_ptr<T>
    {
    public:
        SmartArray(int nb=1):std::shared_ptr<T>(new T[nb], my_array_deleter<T>()),m_nb(nb)
        {
        }
        
        SmartArray(const SmartArray& other):std::shared_ptr<T>(other),m_nb(other.m_nb)
        {
        }

        SmartArray& operator=(const SmartArray& other)
        {
            if(this!=&other)
            {
                this->std::shared_ptr<T>::operator=(other);
                m_nb=other.m_nb;
            }
            return *this;
        }
        
        
        void reallocate(int nb)
        {
            *this=SmartArray<T>(nb);
            m_nb=nb;
        }
        
        SmartArray<T> clone() const
        {
            SmartArray<T> res(m_nb);
            for (int k = 0; k < int(m_nb); ++k)
            {
                res[k]=(*this)[k];
            }
            return res;
        }

        T& operator[](int k) 
        {
            assert(k>=0);
            assert(((unsigned int)k)<m_nb);
            return this->get()[k];
        }

        const T& operator[](int k) const 
        {
            assert(k>=0);
            assert(((unsigned int)k)<m_nb);
            return this->get()[k];
        }
        
        size_t getLength() const
        {
            return m_nb;
        }

        void setValues(const T& t)
        {
            for (unsigned int k = 0; k < m_nb; ++k)
            {
                this->get()[k]=t;
            }
        }
    private:
        size_t m_nb;
    };

    struct ImageSize
    {
        int m_width;
        int m_height;
        int m_nc;

        ImageSize(int width, int height, int nc=1);
    };
    
    class ImageF
    {
        // an image class with float entries,
        // note: row-major and deinterleaved ie im(x,y,c)=im[x+width*y+numpixel*c]. Deinterleaved means one chanel at a time: R1,R2,...,G1,G2,...,B1,B2,..., in comparison to interleaved meaning all channels at once: R1,G1,B1,R2,G2,B2,...
        // note: copy are shallow by default, use clone if you want otherwise
    public:
        ImageF();
        ImageF(const ImageSize& size);
        ImageF(float* buf, const ImageSize& size, bool interleaved=false);
        ImageF(SmartArray<float> buf, const ImageSize& size);
        ImageF(const ImageF& other);
        virtual ~ImageF();

        ImageF& operator=(const ImageF& other);
        ImageF clone() const;

        
        const ImageSize& getSize() const;
        int getNumPixels() const;
        float* getData() const;
        SmartArray<float> getInterleavedData() const;
        float& operator[](int k);
        float operator[](int k) const;

        float& operator()(int x, int y, int c=0);
        float operator()(int x, int y, int c=0) const;

        ImageF operator+(const ImageF& other) const;
        ImageF operator-(const ImageF& other) const;
        ImageF operator*(const ImageF& other) const;

        ImageF pow(float exp) const;

        ImageF transpose() const;
        ImageF crop(int x1, int x2, int y1, int y2, int c1=0, int c2=-1) const;
        ImageF centercrop(int width, int height, int c1=0, int c2=-1) const;

    private:
        SmartArray<float> m_buf;
        ImageSize m_size;
    };
}

#endif /* end of include guard: _IMAGE_H_ */

