#include <cmath>
#include <iostream>
extern "C" {
#include "linterp-src/lprefilt.h"
#include "common/pComponent.h"
}

#include "common/utility.h"
#include "interpolation/linearInterpolation.h"

#ifndef M_PI
/** @brief The constant pi */
#define M_PI    3.14159265358979323846264338327950288
#endif

/** @brief Macro to get the number of elements in a static array */
#define NUMEL(x)    (sizeof(x)/sizeof(*(x)))

using namespace ipol;

LinearInterpolatedImageF::LinearInterpolatedImageF(const std::string& kernelName, const std::string& boundaryName):m_bound(getBoundaryType(boundaryName)),m_kernel(getInterpKernel(kernelName))
{
}

LinearInterpolatedImageF::~LinearInterpolatedImageF()
{
}

float
LinearInterpolatedImageF::operator()(float x, float y, int c) const
{
    float value;
    const ImageSize& size= m_prefiltered.getSize();
    LinInterp2d(&value, m_prefiltered.getData()+c*m_prefiltered.getNumPixels(), size.m_width, size.m_height, &x, &y, 1, m_kernel.m_kernelFunc, m_kernel.m_kernelRadius, m_kernel.m_kernelNormalize, m_bound);
    return value;
}

SmartArray<float>
LinearInterpolatedImageF::operator()(SmartArray<float> xs, SmartArray<float> ys, int c) const
{
    SmartArray<float> values;
    const ImageSize& size= m_prefiltered.getSize();
    int nc=size.m_nc;
    float* inBuf=m_prefiltered.getData();
    if(c>=0)
    {
        inBuf+=c*m_prefiltered.getNumPixels();
        nc=1;
    }
    int numentries= nc*xs.getLength();
    values.reallocate(numentries);
    LinInterp2d(values.get(), inBuf, size.m_width, size.m_height,
                xs.get(), ys.get(), numentries,
                m_kernel.m_kernelFunc, m_kernel.m_kernelRadius, m_kernel.m_kernelNormalize, m_bound);
    return values;
}

ImageF
LinearInterpolatedImageF::onSeparableGrid(SmartArray<float> xs, SmartArray<float> ys, int c) const
{
    const ImageSize& inSize=m_prefiltered.getSize();
    int nc=inSize.m_nc;
    float* inBuf=m_prefiltered.getData();
    if(c>=0)
    {
        inBuf+=c*m_prefiltered.getNumPixels();
        nc=1;
    }
    ImageF out(ImageSize(xs.getLength(), ys.getLength(), nc));
    LinInterpSeparableGrid(out.getData(), xs.get(), xs.getLength(),
                           ys.get(), ys.getLength(),	inBuf,
                           inSize.m_width, inSize.m_height, nc,
                           m_kernel.m_kernelFunc, m_kernel.m_kernelRadius, m_kernel.m_kernelNormalize, m_bound);
    return out;
}


void
LinearInterpolatedImageF::setImage(const ImageF& image)
{
    m_image=image.clone();
    prefilter();
}

void
LinearInterpolatedImageF::prefilter()
{
    m_prefiltered=m_image.clone();
    float* buf=m_prefiltered.getData();
    const ImageSize& size=m_prefiltered.getSize();
    if(m_kernel.m_prefilterNumAlpha)
        PrefilterImage(buf, size.m_width, size.m_height, size.m_nc,
                       m_kernel.m_prefilterAlpha, m_kernel.m_prefilterNumAlpha,
                       m_kernel.m_prefilterScale, m_bound);

}


/////// InterpolationKernel


/**
 * @brief Nearest Neighbor interpolation kernel (KernelRadius = 0.5)
 * @param x coordinate at which to evaluate the kernel
 * @return the kernel value at x
 */
float NearestNeighborKernel(float x)
{
    if(-0.5f <= x && x < 0.5f)
        return 1;
    else
        return 0;
}


/** @brief Bilinear interpolation kernel (KernelRadius = 1) */
static float BilinearKernel(float x)
{
    x = fabs(x);

    if(x < 1)
        return 1 - x;
    else
        return 0;
}


/** @brief Bicubic interpolation kernel (KernelRadius = 2) */
static float BicubicKernel(float x)
{
    const float alpha = -0.5f;

    x = fabs(x);

    if(x < 2)
    {
        if(x <= 1)
            return ((alpha + 2)*x - (alpha + 3))*x*x + 1;
        else
            return ((alpha*x - 5*alpha)*x + 8*alpha)*x - 4*alpha;
    }
    else
        return 0;
}


/** @brief Lanczos-2 interpolation kernel (KernelRadius = 2) */
static float Lanczos2Kernel(float x)
{
    if(-2 < x && x < 2)
    {
        if(x != 0)
            return sin(M_PI*x)*sin((M_PI/2)*x) / ((M_PI*M_PI/2)*x*x);
        else
            return 1;
    }
    else
        return 0;
}


/** @brief Lanczos-3 interpolation kernel (KernelRadius = 3) */
static float Lanczos3Kernel(float x)
{
    if(-3 < x && x < 3)
    {
        if(x != 0)
            return sin(M_PI*x)*sin((M_PI/3)*x) / ((M_PI*M_PI/3)*x*x);
        else
            return 1;
    }
    else
        return 0;
}


/** @brief Lanczos-4 interpolation kernel (KernelRadius = 4) */
static float Lanczos4Kernel(float x)
{
    if(-4 < x && x < 4)
    {
        if(x != 0)
            return sin(M_PI*x)*sin((M_PI/4)*x) / ((M_PI*M_PI/4)*x*x);
        else
            return 1;
    }
    else
        return 0;
}


/** @brief Quadratic B-spline prefilter */
static const float BSpline2Prefilter[1] =
    {-1.715728752538099e-1}; /* exact value: -3 + sqrt(8) */

/** @brief Quadratic B-spline kernel (KernelRadius = 1.5) */
static float BSpline2Kernel(float x)
{
    x = fabs(x);

    if(x <= 0.5f)
        return 0.75f - x*x;
    else if(x < 1.5f)
    {
        x = 1.5f - x;
        return x*x/2;
    }
    else
        return 0;
}


/** @brief Quadratic Schaum kernel (KernelRadius = 1.5) */
static float Schaum2Kernel(float x)
{
    x = fabs(x);

    /* This kernel is discontinuous.  At discontinuous points, it takes the
    average value of the left and right limits. */
    if(x < 0.5f)
        return 1 - x*x;
    else if(x == 0.5f)
        return 0.5625;
    else if(x < 1.5f)
        return (x - 3)*x/2 + 1;
    else if(x == 1.5f)
        return -0.0625;
    else
        return 0;
}


/** @brief Cubic Schaum kernel (KernelRadius = 2) */
static float Schaum3Kernel(float x)
{
    x = fabs(x);

    if(x <= 1)
        return ((x - 2)*x - 1)*x/2 + 1;
    else if(x < 2)
        return ((-x + 6)*x - 11)*x/6 + 1;
    else
        return 0;
}



/** @brief Cubic B-spline prefilter coefficients */
static const float BSpline3Prefilter[1] =
    {-2.679491924311227e-1}; /* exact value: -2 + sqrt(3) */

/** @brief Cubic B-spline kernel (KernelRadius = 2) */
static float BSpline3Kernel(float x)
{
    x = fabs(x);

    if(x < 1)
        return (x/2 - 1)*x*x + 0.66666666666666667f;
    else if(x < 2)
    {
        x = 2 - x;
        return x*x*x/6;
    }
    else
        return 0;
}


/** @brief Quintic B-spline prefilter coefficients */
static const float BSpline5Prefilter[2] =
    {-4.309628820326465e-2, /* exact: sqrt(13*sqrt(105)+135)/sqrt(2)-sqrt(105)/2-13/2.0 */
    -4.305753470999738e-1}; /* exact: sqrt(105)/2+sqrt(135-13*sqrt(105))/sqrt(2)-13/2.0 */

/** @brief Quintic B-spline kernel (KernelRadius = 3) */
static float BSpline5Kernel(float x)
{
    x = fabs(x);

    if(x <= 1)
    {
        float xSqr = x*x;
        return (((-10*x + 30)*xSqr - 60)*xSqr + 66) / 120;
    }
    else if(x < 2)
    {
        x = 2 - x;
        return (1 + (5 + (10 + (10 + (5 - 5*x)*x)*x)*x)*x) / 120;
    }
    else if(x < 3)
    {
        float xSqr;
        x = 3 - x;
        xSqr = x*x;
        return xSqr*xSqr*x / 120;
    }
    else
        return 0;
}


/** @brief Septic B-spline prefilter coefficients */
static const float BSpline7Prefilter[3] =
    {-9.148694809608277e-3, -1.225546151923267e-1, -5.352804307964382e-1};

/** @brief Septic B-spline kernel (KernelRadius = 4) */
static float BSpline7Kernel(float x)
{
    x = fabs(x);

    if(x <= 1)
    {
        float xSqr = x*x;
        return ((((35*x - 140)*xSqr + 560)*xSqr - 1680)*xSqr + 2416) / 5040;
    }
    else if(x <= 2)
    {
        x = 2 - x;
        return (120 + (392 + (504 + (280 + (-84 + (-42 +
            21*x)*x)*x*x)*x)*x)*x) / 5040;
    }
    else if(x < 3)
    {
        x = 3 - x;
        return (((((((-7*x + 7)*x + 21)*x + 35)*x + 35)*x
            + 21)*x + 7)*x + 1) / 5040;
    }
    else if(x < 4)
    {
        float xSqr;
        x = 4 - x;
        xSqr = x*x;
        return xSqr*xSqr*xSqr*x / 5040;
    }
    else
        return 0;
}


/** @brief Nonic B-spline prefilter coefficients */
static const float BSpline9Prefilter[4] =
    {-2.121306903180818e-3, -4.322260854048175e-2,
    -2.017505201931532e-1, -6.079973891686259e-1};

/** @brief Nonic B-spline kernel (KernelRadius = 5) */
static float BSpline9Kernel(float x)
{
    x = fabs(x);

    if(x <= 1)
    {
        float xSqr = x*x;
        return (((((-63*x + 315)*xSqr - 2100)*xSqr + 11970)*xSqr
            - 44100)*xSqr + 78095) / 181440;
    }
    else if(x <= 2)
    {
        x = 2 - x;
        return (14608 + (36414 + (34272 + (11256 + (-4032 + (-4284 + (-672
            + (504 + (252 - 84*x)*x)*x)*x)*x)*x)*x)*x)*x) / 362880;
    }
    else if(x <= 3)
    {
        x = 3 - x;
        return (502 + (2214 + (4248 + (4536 + (2772 + (756 + (-168 + (-216
            + (-72 + 36*x)*x)*x)*x)*x)*x)*x)*x)*x) / 362880;
    }
    else if(x < 4)
    {
        x = 4 - x;
        return (1 + (9 + (36 + (84 + (126 + (126 + (84 + (36 + (9
            - 9*x)*x)*x)*x)*x)*x)*x)*x)*x) / 362880;
    }
    else if(x < 5)
    {
        float xCube;
        x = 5 - x;
        xCube = x*x*x;
        return xCube*xCube*xCube / 362880;
    }
    else
        return 0;
}


/** @brief 11th-Degree B-spline prefilter coefficients */
static const float BSpline11Prefilter[5] =
    {-5.105575344465021e-4, -1.666962736623466e-2, -8.975959979371331e-2,
    -2.721803492947859e-1, -6.612660689007345e-1};

/** @brief 11th-Degree B-spline kernel (KernelRadius = 6) */
static float BSpline11Kernel(float x)
{
    x = fabs(x);

    if(x <= 1)
    {
        float xSqr = x*x;
        return (15724248 + (-7475160 + (1718640 + (-255024 + (27720
            + (-2772 + 462*x)*xSqr)*xSqr)*xSqr)*xSqr)*xSqr) / 39916800;
    }
    else if(x <= 2)
    {
        x = 2 - x;
        return (2203488 + (4480872 + (3273600 + (574200 + (-538560
            + (-299376 + (39600 + (7920 + (-2640 + (-1320
            + 330*x)*x)*x)*x)*x*x)*x)*x)*x)*x)*x) / 39916800;
    }
    else if(x <= 3)
    {
        x = 3 - x;
        return (152637 + (515097 + (748275 + (586575 + (236610 + (12474
            + (-34650 + (-14850 + (-495 + (1485
            + (495-165*x)*x)*x)*x)*x)*x)*x)*x)*x)*x)*x) / 39916800;
    }
    else if(x < 4)
    {
        x = 4 - x;
        return (2036 + (11132 + (27500 + (40260 + (38280 + (24024 + (9240
            + (1320 + (-660 + (-440 + (-110
            + 55*x)*x)*x)*x)*x)*x)*x)*x)*x)*x)*x) / 39916800;
    }
    else if(x < 5)
    {
        x = 5 - x;
        return (1 + (11 + (55 + (165 + (330 + (462 + (462 + (330 + (165
            + (55 + (11 - 11*x)*x)*x)*x)*x)*x)*x)*x)*x)*x)*x) / 39916800;
    }
    else if(x < 6)
    {
        float xSqr, xPow4;
        x = 6 - x;
        xSqr = x*x;
        xPow4 = xSqr*xSqr;
        return xPow4*xPow4*xSqr*x / 39916800;
    }
    else
        return 0;
}


/** @brief Cubic o-Moms prefilter coefficients */
static const float OMoms3Prefilter[1] =
    {-3.441311542550502e-1}; /* exact: (sqrt(105) - 13)/8 */

/** @brief Cubic o-Moms kernel (KernelRadius = 2) */
static float OMoms3Kernel(float x)
{
    x = fabs(x);

    if(x < 1)
        return ((x/2 - 1)*x + 1/14.0f)*x + 13/21.0f;
    else if(x < 2)
        return ((-x/6 + 1)*x - 85/42.0f)*x + 29/21.0f;
    else
        return 0;
}


/** @brief Quintic o-Moms prefilter coefficients */
static const float OMoms5Prefilter[2] =
    {-7.092571896868541e-2, -4.758127100084396e-1};

/** @brief Quintic oMoms kernel (KernelRadius = 3) */
static float OMoms5Kernel(float x)
{
    x = fabs(x);

    if(x <= 1)
        return (((((-10*x + 30)*x - (200/33.0f))*x
            - (540/11.0f))*x - (5/33.0f))*x + (687/11.0)) / 120;
    else if(x < 2)
        return (((((330*x - 2970)*x + 10100)*x
            - 14940)*x + 6755)*x + 2517)/7920;
    else if(x < 3)
    {
        float xSqr;
        x = 3 - x;
        xSqr = x*x;
        return ((xSqr + (20/33.0f))*xSqr + (1/66.0f))*x / 120;
    }
    else
        return 0;
}

/** @brief Septic oMoms prefilter coefficients */
static const float OMoms7Prefilter[3] =
    {-1.976842538386140e-2, -1.557007746773578e-1, -5.685376180022930e-1};

/** @brief Septic oMoms kernel (KernelRadius = 4) */
static float OMoms7Kernel(float x)
{
    x = fabs(x);

    if(x <= 1)
        return (((((((15015*x - 60060)*x + 21021)*x + 180180)*x + 2695)*x
            - 629244)*x + 21)*x + 989636) / 2162160;
    else if(x <= 2)
    {
        x = 2 - x;
        return (x*(x*(x*(x*(x*(x*(5005*x - 10010) - 13013) - 10010) + 54285)
            + 119350) + 106267) + 36606) / 1201200;
    }
    else if(x <= 3)
    {
        x = 3 - x;
        return (x*(x*(x*(x*(x*(x*(-15015*x + 15015) + 24024) + 90090)
            + 102410) + 76230) + 31164) + 5536) / 10810800;
    }
    else if(x < 4)
    {
        float xSqr;
        x = 4 - x;
        xSqr = x*x;
        return (x*(xSqr*(xSqr*(2145*xSqr + 3003) + 385) + 3)) / 10810800;
    }
    else
        return 0;
}


/** @brief Table of all the interpolation methods */
static InterpolationKernel InterpMethodTable[] =
{InterpolationKernel("nearest", NearestNeighborKernel, 0.51f, 0, 0, 0, 1),
 InterpolationKernel("bilinear", BilinearKernel, 1, 0, 0, 0, 1),
 InterpolationKernel("bicubic", BicubicKernel, 2, 0, 0, 0, 1),
 InterpolationKernel("lanczos2", Lanczos2Kernel, 2, 1, 0, 0, 1),
 InterpolationKernel("lanczos3", Lanczos3Kernel, 3, 1, 0, 0, 1),
 InterpolationKernel("lanczos4", Lanczos4Kernel, 4, 1, 0, 0, 1),
 InterpolationKernel("schaum2", Schaum2Kernel, 1.51f, 0, 0, 0, 1),
 InterpolationKernel("schaum3", Schaum3Kernel, 2, 0, 0, 0, 1),
 InterpolationKernel("bspline2", BSpline2Kernel, 1.5f, 0,
                     NUMEL(BSpline2Prefilter), BSpline2Prefilter, 8),
 InterpolationKernel("bspline3", BSpline3Kernel, 2, 0,
                     NUMEL(BSpline3Prefilter), BSpline3Prefilter, 6),
 InterpolationKernel("bspline5", BSpline5Kernel, 3, 0,
                     NUMEL(BSpline5Prefilter), BSpline5Prefilter, 120),
 InterpolationKernel("bspline7", BSpline7Kernel, 4, 0,
                     NUMEL(BSpline7Prefilter), BSpline7Prefilter, 5040),
 InterpolationKernel("bspline9", BSpline9Kernel, 5, 0,
                     NUMEL(BSpline9Prefilter), BSpline9Prefilter, 362880),
 InterpolationKernel("bspline11", BSpline11Kernel, 6, 0,
                     NUMEL(BSpline11Prefilter), BSpline11Prefilter, 39916800),
 InterpolationKernel("omoms3", OMoms3Kernel, 2, 0,
                     NUMEL(OMoms3Prefilter), OMoms3Prefilter, 21/4.0f),
 InterpolationKernel("omoms5", OMoms5Kernel, 3, 0,
                     NUMEL(OMoms5Prefilter), OMoms5Prefilter, 7920/107.0f),
 InterpolationKernel("omoms7", OMoms7Kernel, 4, 0,
                     NUMEL(OMoms7Prefilter), OMoms7Prefilter, 675675/346.0f)};


/**
 * @brief Get the interpmethod struct for an interpolation method by name
 * @param name name of the interpolation method
 * @return pointer to interpmethod struct on success, null pointer on failure.
 *
 * Choices are
 *   - "nearest"
 *   - "bilinear"
 *   - "bicubic"
 *   - "lanczosN" with N = 2, 3, or 4
 *   - "schaumN" with N = 2 or 3
 *   - "bsplineN" with N = 2, 3, 5, 7, 9, or 11
 *   - "omomsN" with N = 3, 5 or 7
 */
InterpolationKernel ipol::getInterpKernel(const std::string& name)
{
    int i;
    for(i = 0; i < (int)NUMEL(InterpMethodTable); i++)
        if(InterpMethodTable[i].m_name== name)
            return InterpMethodTable[i];

    return InterpMethodTable[0];
}


InterpolationKernel::InterpolationKernel(const std::string& name, float (*kernelFunc)(float), float kernelRadius, int kernelNormalize, int prefilterNumAlpha, const float * prefilterAlpha, float prefilterScale):m_name(name), m_kernelFunc(kernelFunc), m_kernelRadius(kernelRadius), m_kernelNormalize(kernelNormalize),m_prefilterNumAlpha(prefilterNumAlpha), m_prefilterAlpha(prefilterAlpha), m_prefilterScale(prefilterScale)
{
}

InterpolationKernel::~InterpolationKernel()
{
}



SincInterpolatedImageF::SincInterpolatedImageF(bool periodic, bool centered, bool pps, int accIter, int axis):
    m_periodic(periodic), m_centered(centered), m_pps(pps), m_accIter(accIter), m_axis(axis)
{
}

float
SincInterpolatedImageF::operator()(float x, float y, int c) const
{
    //!fixme
    float value=x+y+c;
    return value;
}

SmartArray<float> 
SincInterpolatedImageF::operator()(SmartArray<float> xs, SmartArray<float> ys, int c) const
{
    SmartArray<float> values;
    //!fixme
    values=xs; values[c]=ys[c];
    return values;
}

ImageF 
SincInterpolatedImageF::onSeparableGrid(SmartArray<float> xs, SmartArray<float> ys, int c) const
{
    ImageF img=m_image;
    if(m_axis==0 || m_axis==-1)
    {
        int currentAxis=0;
        img=interp1D(img, currentAxis, xs, c);
        if(m_axis==0)
        {
            LinearInterpolatedImageF nearest("nearest", "cst"); 
            nearest.setImage(img);
            int w=img.getSize().m_width;
            img=nearest.onSeparableGrid(makeRange(0, w-1, 1., true), ys, c);
        }
    }
    if(m_axis==1 || m_axis==-1)
    {
        int currentAxis=1;
        img=interp1D(img, currentAxis, ys, c);
        if(m_axis==1)
        {
            LinearInterpolatedImageF nearest("nearest", "cst"); 
            nearest.setImage(img);
            int h=img.getSize().m_height;
            img=nearest.onSeparableGrid(xs, makeRange(0, h-1, 1., true), c);
        }
    }
    return img;
}

ImageF
SincInterpolatedImageF::interp1D(ImageF img, int axis,  SmartArray<float> locs, int c) const
{
    if(axis==1)
    {
        ImageF out= interp1D(img.transpose(), 0, locs, c);
        return out.transpose();
    }
    const ImageSize& inSize=img.getSize();
    int nc=inSize.m_nc;
    int width=inSize.m_width;
    int height=inSize.m_height;
    if(m_pps)
    {
        // compute the periodic+smooth decomposition
        SmartArray<float> imVec=img.getInterleavedData();
        ImageF s(inSize);
        scomponent(s.getData(), imVec.get(), width, height, nc);
        deinterleaveInPlace(s.getData(), width, height, nc);
        ImageF p=img-s;
        // interpolate the periodic component with dft interpolation (sincd)
        bool periodic=true;
        bool pps=false;
        SincInterpolatedImageF dftIm(periodic, pps, m_centered, m_accIter, axis);
        dftIm.setImage(img);
        ImageF tmp=dftIm.interp1D(p, axis, locs, c);
        // interpolate the smooth component with bilinear approach
        LinearInterpolatedImageF intim("bilinear","cst");
        intim.setImage(s);
        SmartArray<float> ylocs=makeRange(0, height-1, 1., true);
        ImageF out=tmp+intim.onSeparableGrid(locs,ylocs,c);
        return out;
    }
    int oWidth=locs.getLength();
    int oHeight=height;
    float* inBuf=img.getData();
    if(c>=0)
    {
        inBuf+=c*img.getNumPixels();
        nc=1;
    }
    float cst=0.;
    if(m_centered)
    {
        for (int k = 0; k < nc*img.getNumPixels(); ++k)
        {
            cst+=img[k];
        }
        cst/=nc*img.getNumPixels();
    }
    ImageF out(ImageSize(oWidth, oHeight, nc));
    SmartArray<double> weights(width);
    SmartArray<double> partSumsLeft;
    SmartArray<double> partSumsRight;
    for (int c=0; c < nc; c++) 
    {
        for(int k=0; k<oWidth; k++)
        {
            float x=locs[k];
            for (int i = 0; i < width; ++i)
            {
                if(m_periodic)
                {
                    weights[i]=sincd(x-i,width);
                }
                else
                {
                    weights[i]=sinc(x-i);
                }
            }
            int i0=round(x-oWidth/2)+oWidth/2;
            if(i0==width)
                i0=width-1;
            else if(i0==-1)
                i0=1;
            partSumsLeft.reallocate(i0+1);
            partSumsRight.reallocate(width-i0-1);
            for (int m = 0; m < height; ++m)
            {
                double val=0;
                for (int i = 0; i <= i0; ++i)
                {
                    int iflipped=i0-i;
                    val+=(inBuf[c*width*height+m*width+iflipped]-cst)*weights[iflipped];
                    partSumsLeft[i]=val;
                }                               
                double val2=val;
                for (int i = i0+1; i < width; ++i)
                {
                    val+=(inBuf[c*width*height+m*width+i]-cst)*weights[i];
                    partSumsRight[i-(i0+1)]=val-val2;
                }
                if(m_accIter>0)
                {
                    val=vanWijngaarden(partSumsLeft,m_accIter) + vanWijngaarden(partSumsRight,m_accIter);
                }
                out[c*oWidth*height+m*oWidth+k]=val+cst;
            }
        }
    }
    return out;
}

double 
SincInterpolatedImageF::sinc(double x)
{
    if(fabs(x)>1e-15)
        return sin(M_PI*x)/(M_PI*x);
    else
        return 1.-M_PI*M_PI*x*x/6.;// taylor series ( obtained thanks to http://www.wolframalpha.com/input/?i=series+sinc%28Pi+x%29+ )
}

double 
SincInterpolatedImageF::sincd(double x, int N)
{
    //!fixme: should be different when N is even
    // take the periodized index corresponding to x or fabs(x)
    x=fabs(x);
    x=fmod((double)x+N/2,(double)N)-N/2;
    if(fabs(x)>1e-15)//&&fabs(x+N)>1e-15&&fabs(x-N)>1e-15)
        return sin(M_PI*x)/N/sin(M_PI*x/N);
    else
        return 1.;
    // We could have used a Taylor series here as well:
    // return 1.-M_PI*M_PI*x*x/6.*(N*N-1)/(N*N);
}

double 
ipol::vanWijngaarden( SmartArray<double> partSums, int itermax)
{
    int len= partSums.getLength();
    if(len==0)
        return 0;
    else if(len==1)
        return partSums[0];
    SmartArray<double> tmp=partSums.clone();
    int last= (len*2)/3;
    if(last>itermax)
        last=itermax;
    for (int j = 0; j <= last; ++j)
    {
           for (int k = 0; k < len-(j+1); ++k)
           {
               tmp[k]=0.5*(tmp[k]+tmp[k+1]);
           }
    }
    return tmp[len-1-last];
}

//////// BoundaryType
boundaryhandling ipol::getBoundaryType(const std::string& boundaryName)
{
    boundaryhandling b= BOUNDARY_CONSTANT;
    if(boundaryName=="hsym")
        b = BOUNDARY_HSYMMETRIC;
    else if(boundaryName=="wsym")
        b = BOUNDARY_WSYMMETRIC;
    return b;
}
