#ifndef _LINEARINTERPOLATIONAPI_H_
#define _LINEARINTERPOLATIONAPI_H_


extern "C"
{
#include "linterp-src/linterp.h"
}

#include "interpolation/interpolationAPI.h"

namespace ipol
{
/** @brief class describing a linear interpolation method */
    class InterpolationKernel
    {
    public:
        InterpolationKernel(const std::string& name, float (*kernelFunc)(float), float kernelRadius, int kernelNormalize, int prefilterNumAlpha, const float * prefilterAlpha, float prefilterScale);
        virtual ~InterpolationKernel();

    public:
        /** @brief Name of the method */
        std::string m_name;
        /** @brief Interpolation kernel or basis function */
        float (*m_kernelFunc)(float);
        /** @brief Support radius of Kernel */
        float m_kernelRadius;
        /** @brief Nonzero value indicates that the kernel should be normalized */
        int m_kernelNormalize;
        /** @brief The number of filter pairs for prefilting */
        int m_prefilterNumAlpha;
        /** @brief Array of prefilter coefficients */
        const float *m_prefilterAlpha;
        /** @brief Constant scale factor to use with prefiltering */
        float m_prefilterScale;
    };

    InterpolationKernel getInterpKernel(const std::string& name);

    class LinearInterpolatedImageF: public InterpolatedImageF 
    {
    public:
        LinearInterpolatedImageF(const std::string& methodName, const std::string& boundaryName);
        virtual ~LinearInterpolatedImageF();

        virtual void setImage(const ImageF& image);

        virtual float operator()(float x, float y, int c=0) const;
                
        // if c==-1 returns all layers and only layer c otherwise
        // xs and ys should have the same length
        virtual SmartArray<float> operator()(SmartArray<float> xs, SmartArray<float> ys, int c=-1) const;

        // if c==-1 returns all layers and only layer c otherwise
        // this is a more efficient implementation than the default one, using 1D ScanLine algorithm
        virtual ImageF onSeparableGrid(SmartArray<float> xs, SmartArray<float> ys, int c=-1) const;

    protected:
        void prefilter();

    protected:
        ImageF m_prefiltered;
        boundaryhandling m_bound;
        InterpolationKernel m_kernel;
    };    

    class SincInterpolatedImageF: public InterpolatedImageF
    {
    public:
        SincInterpolatedImageF(bool periodic, bool centered, bool pps, int accIter, int axis);

        virtual float operator()(float x, float y, int c=0) const;

        virtual SmartArray<float> operator()(SmartArray<float> xs, SmartArray<float> ys, int c=-1) const;

        virtual ImageF onSeparableGrid(SmartArray<float> xs, SmartArray<float> ys, int c=-1) const;

    private:
        static double sinc(double x);
        static double sincd(double x, int N);
        ImageF interp1D(ImageF img, int axis, SmartArray<float> locs, int c=-1) const;
    private:
        bool m_periodic;
        bool m_centered;
        bool m_pps;
        int m_accIter;
        int m_axis;
    };

    double vanWijngaarden( SmartArray<double> partSums, int itermax=6);
    boundaryhandling getBoundaryType(const std::string& boundaryName);

}

#endif /* _LINEARINTERPOLATIONAPI_H_ */

