/**
 * @file lprefilt.c
 * @brief Prefiltering for linear interpolation
 * @author Pascal Getreuer <getreuer@gmail.com>
 *
 * This file implements the prefiltering \c PrefilterImage on the input
 * image needed before performing interpolation needed with B-splines
 * and o-Moms.  Also implemented here is \c PsfPreFilter, the
 * prefiltering adjustment needed when interpolating with a PSF.
 *
 *
 * Copyright (c) 2010-2011, Pascal Getreuer
 * All rights reserved.
 *
 * This program is free software: you can use, modify and/or
 * redistribute it under the terms of the simplified BSD License. You
 * should have received a copy of this license along this program. If
 * not, see <http://www.opensource.org/licenses/bsd-license.html>.
 */

#include <stdlib.h>
#include <math.h>
#include "lprefilt.h"


/**
 * @brief 1D in-place filtering with a first-order recursive filter pair
 * @param Data pointer to data to be filtered
 * @param Stride stride between successive elements
 * @param N number of samples
 * @param alpha filter coefficient
 * @param Boundary the kind of boundary handling to use
 *
 * Applies the causal recursive filter
 *     1/(1 - alpha z^-1)
 * followed by the anti-causal recursive filter
 *     -alpha/(1 - alpha z).
 * The coefficient alpha must satisify |alpha| < 1 for stability.
 *
 * With respect to boundary handling, filtering is computed with relative
 * accuracy Eps = 1e-4 for half- and whole-sample symmetric boundaries and it
 * is exact for constant extension.  Note, however, that for constant extension
 * the infinite grid result is not exactly constant beyond the boundaries
 * (rather it decays to constant).
 */
static void PrefilterScan(float *Data, int Stride, int N, float alpha,
    boundaryhandling Boundary)
{
    const float Eps = 1e-4;
    float Sum, Weight, Last;
    int i, iEnd, n0;

    n0 = (int)ceil(log(Eps)/log(fabs(alpha)));

    if(n0 > N)
        n0 = N;

    switch(Boundary)
    {
        case BOUNDARY_CONSTANT:
            Sum = Data[0]/(1 - alpha);
            break;
        case BOUNDARY_WSYMMETRIC:
            Sum = Data[0];
            Weight = 1;
            iEnd = n0*Stride;

            for(i = Stride; i < iEnd; i += Stride)
            {
                Weight *= alpha;
                Sum += Data[i]*Weight;
            }
            break;
        default: /* BOUNDARY_HSYMMETRIC */
            Sum = Data[0]*(1 + alpha);
            Weight = alpha;
            iEnd = n0*Stride;

            for(i = Stride; i < iEnd; i += Stride)
            {
                Weight *= alpha;
                Sum += Data[i]*Weight;
            }
            break;
    }

    Last = Data[0] = Sum;
    iEnd = (N - 1)*Stride;

    for(i = Stride; i < iEnd; i += Stride)
    {
        Data[i] += alpha*Last;
        Last = Data[i];
    }

    switch(Boundary)
    {
        case BOUNDARY_CONSTANT:
            Last = Data[iEnd] = (alpha*(-Data[iEnd] + (alpha - 1)*alpha*Last))
                /((alpha - 1)*(alpha*alpha - 1));
            break;
        case BOUNDARY_HSYMMETRIC:
            Data[iEnd] += alpha*Last;
            Last = Data[iEnd] *= alpha/(alpha - 1);
            break;
        case BOUNDARY_WSYMMETRIC:
            Data[iEnd] += alpha*Last;
            Last = Data[iEnd] = (alpha/(alpha*alpha - 1))
                * ( Data[iEnd] + alpha*Data[iEnd - Stride] );
            break;
    }

    for(i = iEnd - Stride; i >= 0; i -= Stride)
    {
        Data[i] = alpha*(Last - Data[i]);
        Last = Data[i];
    }
}


/**
 * @brief Apply a cascade of first-order recursive filter pairs to an image
 * @param Data the image data
 * @param Width, Height, NumChannels image dimensions
 * @param alpha array of alpha values
 * @param NumFilterPairs the number of filter pairs
 * @param ConstantFactor constant multiplicative factor to apply
 * @param Boundary the kind of boundary handling to use
 */
void PrefilterImage(float *Data, int Width, int Height, int NumChannels,
    const float *alpha, int NumFilterPairs, float ConstantFactor,
    boundaryhandling Boundary)
{
    const int NumPixels = Width*Height;
    int k, x, y, Channel;


    /* Square the ConstantFactor for two spatial dimensions */
    ConstantFactor = ConstantFactor*ConstantFactor;

    for(Channel = 0; Channel < NumChannels; Channel++)
    {
        for(x = 0; x < Width; x++)
            for(k = 0; k < NumFilterPairs; k++)
                PrefilterScan(Data + x, Width, Height, alpha[k], Boundary);

        for(y = 0; y < Height; y++)
            for(k = 0; k < NumFilterPairs; k++)
                PrefilterScan(Data + Width*y, 1, Width, alpha[k], Boundary);

        for(k = 0; k < NumPixels; k++)
            Data[k] *= ConstantFactor;

        Data += NumPixels;
    }
}


