#include <cstdlib>
#include <string>
#include <iostream>
#include <fstream>
#include <algorithm>
#include <cmath>
extern "C"
{
#include "iio.h"
#include <math.h>
#include <fftw3.h>
}
#include "interpolation/interpolationAPI.h"
inline double cabs(fftw_complex z)
{
    return sqrt(pow(z[0],2)+pow(z[1],2));
}
using namespace ipol;

SmartArray<double>
computePeriodogram(ImageF im, int x1=0, int x2=-1)
{
    ImageSize size=im.getSize();
    int w=size.m_width, h=size.m_height, nc=size.m_nc;
    if(x2<0)
        x2=w;
    w=(x2-x1);
    int n=w/2+1;
    SmartArray<double> periodogram(n);
    periodogram.setValues(0);
    // Memory allocation
    double *line = static_cast<double*>(fftw_malloc(sizeof(double) * w));
    fftw_complex *dft = static_cast<fftw_complex*>(fftw_malloc(sizeof(fftw_complex) * n));
    // Plans computation
    fftw_plan plan= fftw_plan_dft_r2c_1d(w, line, dft, FFTW_ESTIMATE);

    // Compute the output image row by row :
    for (int y = 0; y < h; y++)
        for(int c = 0; c<nc; c++)
    {
        // Copy and duplicate the current row, and compute the mean value of the line
        double mean=0.;
        for (int x=0; x<w; x++)
        {
            line[x] = im(x1+x,y,c);
            mean+=line[x];
        }
        mean/=w;
        // Remove the mean value to the row
        for (int x=0; x<w; x++)
        {
            line[x]-=mean;
        }
        // Compute the DFT
        fftw_execute(plan);

        // compute the normalized periodogram (normalized so that sum_i periodogram[i]=variance of im)
        for (int i=0; i<n; i++)
        {
            periodogram[i]+=pow(cabs(dft[i]), 2.)/w;
        }
    }
    // normalize the periodogram
    for (int i=0; i<n; i++)
        periodogram[i]/=h*nc;

    // Free memory
    fftw_destroy_plan(plan);
    fftw_free(line);
    fftw_free(dft);

    return periodogram;
}

SmartArray<double> cumulativeMax(SmartArray<double> periodogram)
{
    int n=periodogram.getLength();
    SmartArray<double> cumMaxs(n);
    double cumMax=0;
    for(int i=n-1;i>=0;i--)
    {
        if(cumMax<periodogram[i])
            cumMax=periodogram[i];
        cumMaxs[i]=cumMax;
    }
    return cumMaxs;
}

SmartArray<double> computeSigmas(SmartArray<double> periodogram, SmartArray<double> cumMaxs)
{
    int nbFreqs=periodogram.getLength();
    SmartArray<double> sigAlphas(nbFreqs);
    for (int k = 0; k < nbFreqs; ++k)
    {
        // the goal here is to approximate sigAlpha^2:=1/(2\pi)\int_{[omega|<alpha\pi} 4/abs(1+exp(i*omega))^2 dPsi_{OS}(omega)
        // where alpha=k/nbFreqs and Psi_{OS} is an oversampled periodogram obtained after removing the white-noise component
        // note that 4/abs(1+exp(i*omega))^2 = 2/(1+cos(omega))
        // note also that sigAlpha^2 can be interpreted as an average value
        float sigAlpha2=0.;
        
        for(int j = 0; j < nbFreqs; ++j)
        {
            double omega=j/static_cast<float>(nbFreqs)*M_PI;
            double psiOSj=std::max(0.,periodogram[j]-cumMaxs[k]);
            sigAlpha2+=2./(1+std::cos(omega))*psiOSj;
        }
        sigAlpha2/=static_cast<float>(nbFreqs);
        sigAlphas[k]= std::sqrt(sigAlpha2);
        
    }
    return sigAlphas;
}

SmartArray<float> computeErrorUpperBound(ImageF im,  SmartArray<double> cumMaxs, SmartArray<double> sigAlphas, SmartArray<float> positions, int halfWidth, const std::string& interp="shannon")
{
    int nbPos=positions.getLength();
    int nbFreqs=sigAlphas.getLength();
    SmartArray<float> errs(nbPos);
    // now compute the upper bounds for every position
    for (int i=0; i<nbPos; ++i)
    {
        float pos=positions[i];
        float shortDist=halfWidth+0.5-pos, longDist=halfWidth+0.5+pos;
        float err2=1e15;
        for(int k = 0; k < nbFreqs; ++k)
        {
            float errOS= sigAlphas[k] * (1./(shortDist+0.5) + 1./(longDist+0.5));
            float  errWN2= cumMaxs[k] * (1./shortDist + 1./longDist);
            if(interp=="dft")
            {
              errOS*=2;
              errWN2*=2;
            }
            float err2Candidate=pow(errOS, 2.) + errWN2;
            if(err2>err2Candidate)
                err2=err2Candidate;
        }
        if(interp=="shannon")
        {
            float average=0.;
            int N=im.getNumPixels()*im.getSize().m_nc;
            for (int i = 0; i < N; ++i)
            {
                average+=im[i]/N;
            }
            err2+=pow( average, 2.) * pow(1./(shortDist+0.5) + 1./(longDist+0.5), 2. );
        }
        // do not forget the modulation term
        errs[i]=std::sqrt(err2)*std::abs(std::sin(M_PI*pos))/M_PI;
    }      
    return errs;
}


int main(int argc, char *argv[])
{
    // check command line arguments
    if (argc < 5 || argc > 7 ) {
        std::cerr<<"usage:\n\t"<<argv[0]<<" method step x1 x2 [in [out]]"<<std::endl;
        //                          0         1     2    3  4   5   6  
        return EXIT_FAILURE;
    }
    // parse the aguments
    std::string methodName=argv[1];
    float step=atof(argv[2]);
    float x1=atof(argv[3]);
    float x2=atof(argv[4]);
    std::string inName= (argc>5)? argv[5] : "-";
    std::string outDir= (argc>6)? argv[6] : "./";
    std::string outName=outDir+"/"+methodName+"UB.data";
    // read input image "im"
    int w,h,pd;
    float *buf = static_cast<float*>(iio_read_image_float_split(inName.c_str(), &w, &h, &pd));
    ImageSize size(w,h,pd);
    ImageF im(buf, size);
    // Compute periodogram and the error bounds
    int halfWidth=static_cast<int>(x2-x1) / 2;
    SmartArray<float> positions=makeRange(0,halfWidth,step);
    SmartArray<double> periodogram = computePeriodogram(im,x1,x2);
    SmartArray<double> cumMaxs = cumulativeMax(periodogram);
    SmartArray<double> sigAlphas = computeSigmas(periodogram, cumMaxs);
    SmartArray<float> errUBs=computeErrorUpperBound(im, cumMaxs, sigAlphas, positions,  halfWidth, methodName);
    //dump the error bound entries in a file
    std::ofstream sout(outName.c_str());
    for(unsigned int i=0; i<errUBs.getLength(); i++)
    {
        float shortDist=halfWidth+0.5-positions[i];
        sout<<shortDist<<" "<<errUBs[i]<<std::endl;
    }
    sout.close();
    //dump the periodogram entries in a file if the shannon's method
    if(methodName=="shannon")
    {
        std::string spectrumName=outDir+"/"+"spectrum.data";
        std::ofstream sout(spectrumName.c_str());
        for(unsigned int i=0; i<periodogram.getLength(); i++)
        {
            float alpha=i/(float)periodogram.getLength();
            sout<<alpha<<" "<<periodogram[i]<<std::endl;
        }
        sout.close();
    }
    // free the memory allocated by iio
    free(buf);
    return 0;
}
