#include <assert.h>
#include "iio.h"
#include <math.h>
#include <complex.h>
#include <fftw3.h>

#ifndef M_PI
#define M_PI		3.14159265358979323846	/* pi */
#endif

#include "xmalloc.c"
#define FORI(n) for(int i=0;i<(n);i++)
#define FORL(n) for(int l=0;l<(n);l++)

#ifndef USE_WISDOM
void evoke_wisdom(void) {}
void bequeath_wisdom(void) {}
#else//USE_WISDOM
#include "fftwisdom.c"
#endif//USE_WISDOM


// wrapper around FFTW3 that computes the complex-valued Fourier transform
// of a real-valued image
static void fft_2dfloat(fftwf_complex *fx, float *x, int w, int h)
{
    fftwf_complex *a = fftwf_malloc(w*h*sizeof*a);

    //fprintf(stderr, "planning...\n");
    evoke_wisdom();
    fftwf_plan p = fftwf_plan_dft_2d(h, w, a, fx,
                                     FFTW_FORWARD, FFTW_ESTIMATE);
    bequeath_wisdom();
    //fprintf(stderr, "...planned!\n");

    FORI(w*h) a[i] = x[i]; // complex assignment!
    fftwf_execute(p);

    fftwf_destroy_plan(p);
    fftwf_free(a);
    fftwf_cleanup();
}


// Wrapper around FFTW3 that computes the real-valued inverse Fourier transform
// of a complex-valued frequantial image.
// The input data must be hermitic.
static void ifft_2dfloat(float *ifx,  fftwf_complex *fx, int w, int h)
{
    fftwf_complex *a = fftwf_malloc(w*h*sizeof*a);
    fftwf_complex *b = fftwf_malloc(w*h*sizeof*b);

    //fprintf(stderr, "planning...\n");
    evoke_wisdom();
    fftwf_plan p = fftwf_plan_dft_2d(h, w, a, b,
                                     FFTW_BACKWARD, FFTW_ESTIMATE);
    bequeath_wisdom();
    //fprintf(stderr, "...planned!\n");

    FORI(w*h) a[i] = fx[i];
    fftwf_execute(p);
    float scale = 1.0/(w*h);
    FORI(w*h) {
        fftwf_complex z = b[i] * scale;
        ifx[i] = crealf(z);
        //assert(cimagf(z) < 0.001);
    }
    fftwf_destroy_plan(p);
    fftwf_free(a);
    fftwf_free(b);
    fftwf_cleanup();
}

// if it finds any strange number, sets it to zero
static void normalize_float_array_inplace(float *x, int n)
{
    for (int i = 0; i < n; i++)
        if (!isnormal(x[i]))
            x[i] = 0;

}


static void fft_direct(float *y, float *x, int w, int h, int pd)
{
    float *c = xmalloc(w*h*sizeof*c);
    fftwf_complex *gc = xmalloc(w*h*sizeof*gc);
    FORL(pd) {
        FORI(w*h)
            c[i] = x[i*pd + l];
        fft_2dfloat(gc, c, w, h);
        FORI(w*h) {
            y[2*(i*pd + l)+0] = crealf(gc[i]);
            y[2*(i*pd + l)+1] = cimagf(gc[i]);
        }
    }
    free(c);
    free(gc);
}

static void fft_inverse(float *y, float *x, int w, int h, int pd)
{
    int pdh = pd/2;
    assert(pd == 2*pdh);
    fftwf_complex *c = xmalloc(w*h*sizeof*c);
    float *gc = xmalloc(w*h*sizeof*gc);
    FORL(pdh) {
        FORI(w*h)
            c[i] = x[i*pd + 2*l] + I * x[i*pd+2*l+1];
        ifft_2dfloat(gc, c, w, h);
        FORI(w*h)
            y[i*pdh + l] = gc[i];
    }
    free(c);
    free(gc);
}

static void jumps(float *out, float *in, int w, int h, int pd)
{
    float (*x)[w][pd] = (void*)in;
    float (*y)[w][pd] = (void*)out;
    for (int j = 0; j < h; j++)
        for (int i = 0; i < w; i++)
            for (int l = 0; l < pd; l++)
            {
                y[j][i][l]=0;
            }
    for (int j = 0; j < h; j++)
        for (int l = 0; l < pd; l++)
        {
            y[j][0][l]=x[j][0][l]-x[j][w-1][l];
            y[j][w-1][l]=-y[j][0][l];
        }
    for (int i = 0; i < w; i++)
        for (int l = 0; l < pd; l++)
        {
            y[0][i][l]+=x[0][i][l]-x[h-1][i][l];
            y[h-1][i][l]-=x[0][i][l]-x[h-1][i][l];
        }
}

void scomponent(float *out, float *in, int w, int h, int pd)
{
    float *v=xmalloc(w*h*pd*sizeof*v);
    float *vhat=xmalloc(w*h*2*pd*sizeof*vhat);
    float *shat=xmalloc(w*h*2*pd*sizeof*shat);
    jumps(v, in, w, h, pd);
    float (*x)[w][2*pd] = (void*)vhat;
    float (*y)[w][2*pd] = (void*)shat;
    fft_direct(vhat, v, w, h, pd);
    for (int j = 0; j < h; j++)
        for (int i = 0; i < w; i++)
            for (int l = 0; l < 2*pd; l++)
            {
                y[j][i][l]=x[j][i][l]/(4-2*cos(2*M_PI*j/(float)h)-2*cos(2*M_PI*i/(float)w));
            }    
    normalize_float_array_inplace(shat, w*h*2*pd);
    fft_inverse(out, shat, w, h, 2*pd);
    free(v);
    free(vhat);
    free(shat);
}

void pcomponent(float *out, float *in, int w, int h, int pd)
{
    float *s=xmalloc(w*h*pd*sizeof*s);  
    scomponent(s, in, w, h, pd);
    float (*x)[w][pd] = (void*)in;
    float (*y)[w][pd] = (void*)s;
    float (*z)[w][pd] = (void*)out;
    for (int j = 0; j < h; j++)
        for (int i = 0; i < w; i++)
            for (int l = 0; l < pd; l++)
            {
                z[j][i][l]=x[j][i][l]-y[j][i][l];
            }    
    free(s);
}
