#include <cstdlib>
#include <string>
#include <iostream>
#include <cmath>
extern "C"
{
#include "iio.h"
}
#include "common/utility.h"
#include "interpolation/interpolationAPI.h"
#include "interpolation/linearInterpolation.h"

using namespace ipol;

float parseAngle(const std::string& code)
{
    std::size_t pos=code.find_first_of(':',0);
    if(pos==std::string::npos)
        return 0;
    std::string unit=code.substr(0,pos);
    float value=atof(code.substr(pos+1).c_str());
    if(unit=="deg" or unit=="degree")
        return value*M_PI/180;
    else if(unit=="rad" or unit=="radian")
        return value;
    else if(unit=="round")
        return value*2*M_PI;
    else if(unit=="half")
        return value*M_PI;
    else if(unit=="quarter")
        return value*M_PI/2;
    else
        return 0;
}

int main(int argc, char *argv[])
{
    // check command line arguments
    if (argc < 5 || argc > 7 ) {
        std::cerr<<"usage:\n\t"<<argv[0]<<" method variant angle nbsteps [in [out]]"<<std::endl;
        //                          0          1      2      3      4      5   6 
        return EXIT_FAILURE;
    }
    else
    {
        std::cerr<<argv[0]<<" "<<argv[1]<<" "<<argv[2]<<" "<<argv[3]<<" "<<argv[4]<<std::endl; 
    }
    std::string methodName=argv[1];
    std::string variant=argv[2];
    float angle=parseAngle(argv[3]);
    int nbsteps=atoi(argv[4]);
    std::string inName= (argc>5)? argv[5] : "-";
    std::string outName= (argc>6)? argv[6] : "-";
    std::cout << "rotating image " << inName << " by " << angle << " deg, in " << nbsteps << " steps." << std::endl;
    // read input image "im"
    int w,h,pd;
    float *buf = static_cast<float*>(iio_read_image_float_split(inName.c_str(), &w, &h, &pd));
    ImageSize size(w,h,pd);
    bool interleaved=false; // take care that iio returns an interleaved buffer while ImageF requires a de-interleaved one.
    ImageF im(buf, size, interleaved);
    free(buf);
    InterpolatedImageF* intim;
    if(methodName=="shannon")
    {   
        bool periodic=(variant=="periodic");
        bool centered=(variant=="centered")||(variant=="acc");
        bool pps=(variant=="pps");
        int accIter=0;
        int axis=-1;
        //!fixme: there seems to be no gain for acc>1
        //!fixme: Note that acc=1, probably corresponds to taking Cesaro sum instead of direct summation
        if(variant=="acc")
            accIter=1;
        intim=new SincInterpolatedImageF(periodic, centered, pps, accIter, axis);
    }
    else
    {
        intim= new LinearInterpolatedImageF(methodName,variant);
    }
    // produce the interpolation grid
    SmartArray<float> xs;
    SmartArray<float> ys;
    ImageSize outSize(0,0);
    int outWidth=w, outHeight=h;
    for (int k = 0; k < nbsteps; ++k)
    {
        makeScaleRotationGrid(xs, ys, outWidth, outHeight, outWidth, outHeight, 1, angle/nbsteps);
        intim->setImage(im);
        // save output image
        SmartArray<float> outbuf=(*intim)(xs,ys, -1);
        outSize = ImageSize (outWidth, outHeight, pd);
        im= ImageF(outbuf, outSize);
    }
    delete intim;
    im=im.centercrop(w,h);
    outSize=im.getSize();
    int out_w=outSize.m_width, out_h=outSize.m_height;
    int out_pd=outSize.m_nc;
    iio_save_image_float_split(const_cast<char*>(outName.c_str()), im.getData(), out_w, out_h, out_pd);
    return 0;
}

