#!/usr/bin/env python
import os
import numpy as np
import sys
from distutils.spawn import find_executable
from TDMAsolver import TDMAsolver
display_prog = "display"
#convert_prog = 'inkscape --without-gui --export-pdf='
convert_prog = 'cupsfilter {input} > {output};\
gs -sDEVICE=pdfwrite -dNOPAUSE -dBATCH -dSAFER \
       -dFirstPage=1 -dLastPage=1 \
       -sOutputFile=tmp.pdf {output};\
pdfcrop --resolution 72 tmp.pdf {output};\
       rm tmp.pdf'

if find_executable('cupsfilter') is None:
  convert_prog='convert {input} {output}'
if sys.platform == "darwin":
  os.environ['PATH']+=':/Applications/Inkscape.app/Contents/Resources/bin/'

class Scene:
  def __init__(self,name="svg",height=400,width=400, pathToJS=None):
    self.name = name
    self.items = []
    self.height = height
    self.width = width
    self.defs=[]
    if pathToJS is None:
      self.pathToJS=os.path.abspath('SVGPan.js')
    else:
      self.pathToJS=pathToJS
    return

  def addDef(self, adef):
    self.defs.append(adef)
    return

  def addItem(self,item): self.items.append(item)

  def add(self, elemtype, text='', kwords={}, stylekwords={}):
    #!fixme: Not sure if we really need to separate kwords from stylekwords
    self.addItem(Element(elemtype, text, kwords, stylekwords))
             
  def strarray(self):
    hmargin=100
    var = ["<?xml version=\"1.0\"?>\n",
         "<svg xmlns=\"http://www.w3.org/2000/svg\" xmlns:xlink=\"http://www.w3.org/1999/xlink\" height=\"%d\" width=\"%d\" >\n" % (self.height,self.width+2*hmargin),
         " <script xlink:href=\"%s\"/>\n" %self.pathToJS, 
         " <defs>\n"]+ [line+'\n' for d in self.defs for line in
             d.split('\n')]+[ " </defs>\n",
                 " <g transform=\"matrix(0.9, 0, 0, 0.9, {tx}, 10)\" style=\"fill-opacity:1.0; stroke:black;\n".format(tx=10+hmargin),
         "  stroke-width:1;\">\n"]
    for item in self.items: var += item.strarray()        
    var += [" </g>\n</svg>\n"]
    return var

  def write_svg(self,filename=None):
    if filename:
      self.svgname = filename
    else:
      self.svgname = self.name + ".svg"
    file = open(self.svgname,'w')
    file.writelines(self.strarray())
    file.close()
    return

  def display(self,prog=display_prog):
    os.system("%s %s" % (prog,self.svgname))

  def convert(self, prog=convert_prog, ext='pdf'):
    os.system(prog.format(input=self.svgname, output=os.path.splitext(self.svgname)[0]+'.'+ext))

class Element:
  def __init__(self, elemtype, text='', kwords={}, stylekwords={}):
    self.elemtype=elemtype
    self.text=text
    self.kwords=kwords
    self.stylekwords=stylekwords
    

  def strarray(self):
    start=['  <%s ' % self.elemtype ] +\
         ['%s="%s" ' % (key.replace('_','-'),str(self.kwords[key])) for  key in self.kwords]
    if self.stylekwords!={}:
      start+=['style="']+\
            ['%s:%s;' % (key.replace('_','-'),str(self.stylekwords[key])) for  key in self.stylekwords]+\
            ['"']
    if self.text=='':
      return start+['/>\n']
    else:
      return start+['>\n  %s\n  </%s>\n' %\
                (self.text,self.elemtype)]




def computeSplineControlPoints(pts):
  """
  Return the de Boor Control points that are involved in the following parametrisation of the spline:
  S(t)=    (1-t)^3*P0 + 3*(1-t)^2*t*CP0 +3*(1-t)*t^2*CP1+t^3*P1
  
  This the parametrization such that: S'(0)=3*(CP0-P0) and S'(1)=-3*(CP1-P1)
  Indeed
  S'(t)=3*( -(1-t)^2*P0 + (1-t)(-2*t+(1-t))*CP0 + t*(2*(1-t)-t)*CP1 + t^2*P1 )
      =3*( -(1-t)^2*P0 +    (1-t)(1-3*t)*CP0 +    t*(2-3*t)*CP1 + t^2*P1 )
  We can also compute
  S''(t)=3*( 2*(1-t)*P0 - ((1-3*t)+3*(1-t))*CP0 + ((2-3*t)-3*t)*CP1 + 2*t*P1 )
      =6*(  (1-t)*P0 -        (2-3*t)*CP0 +     (1-3*t)*CP1 +  t*P1 )
  """
  pts2=pts[::-1]
  n=len(pts)-1
  a=np.zeros(n-1)
  b=np.zeros(n)
  c=np.zeros(n-1)
  rx=np.zeros(n)
  ry=np.zeros(n)
  ##left most segment
  b[0]=2
  c[0]=1
  rx[0] = pts[0][0]+2*pts[1][0]
  ry[0] = pts[0][1]+2*pts[1][1]
	##internal segments
  for i in xrange(1,n-1):
		a[i-1]=1
		b[i]=4
		c[i]=1
		rx[i] = 4 * pts[i][0] + 2 * pts[i+1][0]
		ry[i] = 4 * pts[i][1] + 2 * pts[i+1][1]
	##right most segment
  a[n-2]=2
  b[n-1]=7
  rx[n-1] = 8*pts[n-1][0]+pts[n][0]
  ry[n-1] = 8*pts[n-1][1]+pts[n][1]
  ## solve the system
  cpts1=TDMAsolver(a,b,c,np.array([rx,ry]).T)
  cpts1=np.array(cpts1)
  ## Compute the second control points (as mirror of the first
  cpts2=np.concatenate( (2*pts[1:-1]-cpts1[1:], [0.5*(pts[-1]+cpts1[-1])] ))
  return cpts1,cpts2


class SVGPlot:
  fontratio=0.6 # this is the width to height ratio of Courier font (won't work
              # properly for other fonts) 
  def __init__(self, scene):
    self.scene=scene
    self.width=scene.width
    self.height=scene.height
    self.setLims()
    self.plots=[]
    self.debugMesgs=[]
    self.markerTemplate='\n'.join([
         "   <marker id='{name}' stroke='{stroke}' stroke-width='1' fill='{fill}'",
         "     viewBox='-5 -5 10 10' refX='{refX}' refY='{refY}' ",
         "     markerUnits='strokeWidth'",
         "     markerWidth='{markersize}' markerHeight='{markersize}'",
         "     orient='auto'>",
         "    {shape}",     
         "   </marker>"])
    self.markers={}
    # define the marker shapes
    self.markerShapes={}
    self.markerShapes['Triangle']= "<path d='M -4 -4 L 4 0 L -4 4 z' />"
    self.markerShapes['Circle']= "<circle x='-2' y='-2' r='4'/>"
    self.markerShapes['Square']= "<rect x='-4' y='-4' width='8' height='8'/>"
    self.markerShapes['Cross']= "<path d='M -5 -5 L 5 5 M 5 -5 L -5 5 z' />"

  def addMarker(self, name, stroke='none', fill='none', markersize=1, refX=0,
      refY=0):
    key='_'.join((name, stroke, fill, '%s'%markersize, '%s'%refX, '%s'%refY))
    if self.markers.has_key(key):
      return key
    self.markers[key]=self.markerTemplate.format(name=key, stroke=stroke,
      fill=fill, markersize=markersize, refX=refX, refY=refY, shape=self.markerShapes[name]) 
    self.scene.addDef(self.markers[key])
    return key
    
  @classmethod
  def colorstr(cls,rgb):
    if rgb==None:           
      return 'none'
    elif type(rgb)==type(''):
      return rgb
    else:
      return "#%x%x%x" % (rgb[0]/16,rgb[1]/16,rgb[2]/16)
   
  def makeLine(self, x1, y1, x2, y2, stroke='black', stroke_width=3, **kwargs):
    self.scene.add('line','',
           dict(x1=x1,y1=y1,x2=x2,y2=y2),
           dict(kwargs, stroke=self.colorstr(stroke),
              stroke_width=stroke_width))
   
  def makeRectangle(self, x, y, width, height, fill=None, stroke='black', stroke_width=3, **kwargs): 
    self.scene.add('rect','',
           dict(x=x,y=y,height=height,width=width),
           dict(kwargs, fill=self.colorstr(fill),
              stroke=self.colorstr(stroke),
              stroke_width=stroke_width))
   
  def makeCircle(self, cx, cy, r, fill=None, stroke='black', stroke_width=3, **kwargs):
    self.scene.add('circle','',
           dict(cx=cx,cy=cy, r=r),
           dict(kwargs, fill=self.colorstr(fill),
              stroke=self.colorstr(stroke),
              stroke_width=stroke_width))

  @classmethod
  def textToLines(cls, text):
    if type(text)==list:
      return [l for line in text for l in line.split('\n')]
    else:
      return text.split('\n')

  @classmethod
  def computeTextDim(cls, text, returnonly=None):
    lines=cls.textToLines(text)
    height=len(lines)
    width=np.max([len(line) for line in lines])
    if returnonly in ['w','width']:
      return width
    elif returnonly in ['h','height']:
      return height
    else:
      return (width, height)

  @classmethod
  def computeTextWidth(cls, text):
    return cls.computeTextDim(text, returnonly='w')

  @classmethod
  def computeTextHeight(cls, text):
    return cls.computeTextDim(text, returnonly='h')


  def makeText(self, text, x=0, y=0, font_size=12, fill='black',
      text_anchor='start', dominant_baseline='no-change', bgdkwargs=None, **kwargs):
    #!note: hardcode the font family to enforce monospace
    font_family='Courier'
    if kwargs.has_key('font_family'):
      del kwargs['font_family']
    if bgdkwargs is not None:
      dx,dy=0,font_size
      tw,th=self.computeTextDim(text)
      tw,th=tw*(font_size+2)*self.fontratio, th*(font_size+2)
      if text_anchor == 'middle':
        dx=tw/2.
      elif text_anchor == 'end':
        dx=tw
      if dominant_baseline=='central':
        dy=font_size/2.
      self.makeRectangle(x=x-dx-1, y=y-dy-1, width=tw, height=th, **bgdkwargs)
    dy=0
    for line in self.textToLines(text):
      self.scene.add('text',line,
          dict(x=x,y=y+dy, font_size=font_size, font_family=font_family,
                fill=self.colorstr(fill), text_anchor=text_anchor,
                dominant_baseline=dominant_baseline),
             dict(kwargs))
      dy+=font_size+2
   
  def makePath(self, d, fill, stroke, stroke_width, closed=False, **kwargs):
    if closed: d=d+'z'
    self.scene.add('path', '',
           dict(d=d),
           dict(kwargs, fill=self.colorstr(fill),
              stroke=self.colorstr(stroke),
              stroke_width=stroke_width))
   
  def makeImage(self, href, x, y, width, height, **kwargs):
    elem=Element('image', '',
             dict(x=x,y=y,width=width,height=height,**{'xlink:href':href}),
             kwargs)
    self.scene.addItem(elem)

  def getMarkerURL(self, marker, markersize=1, stroke='black', refX=0, refY=0):
    if marker in ['t', 'triangle']:
      return "url(#%s)" %(self.addMarker('Triangle', stroke=stroke,
        markersize=markersize, refX=refX, refY=refY),)
    elif marker in ['s', 'square']:
      return "url(#%s)" %(self.addMarker('Square', stroke=stroke,
        markersize=markersize, refX=refX, refY=refY),)
    elif marker in ['c', 'circle']:
      return "url(#%s)" %(self.addMarker('Circle', stroke=stroke,
        markersize=markersize, refX=refX, refY=refY),)
    if marker in ['ft', 'filled_triangle']:
      return "url(#%s)" %(self.addMarker('Triangle', stroke=stroke,
        fill=stroke, markersize=markersize, refX=refX, refY=refY),)
    elif marker in ['fs', 'filled_square']:
      return "url(#%s)" %(self.addMarker('Square', stroke=stroke,
        fill=stroke, markersize=markersize, refX=refX, refY=refY),)
    elif marker in ['fc', 'filled_circle']:
      return "url(#%s)" %(self.addMarker('Circle', stroke=stroke,
        fill=stroke, markersize=markersize, refX=refX, refY=refY),)
    elif marker in ['x', 'cross']:
      return "url(#%s)" %(self.addMarker('Cross', stroke=stroke,
        markersize=markersize, refX=refX, refY=refY),)
    else:
      return "none"

  def plot(self, pts, stroke='red', stroke_width=3, fill=None, fill_opacity=1, interp='linear', legend=None,
      font_size=16, font_color='white', marker=None,
      markersize=3, xlim=None, ylim=None, drawaxis=False):
    self.setLims(xlim=xlim, ylim=ylim)
    pts=np.array(pts)
    closingstr=''
    if fill is not None:
      p1,p2=[self.rescale(x=x,y=y) for x,y in [(pts[-1,0],0), (pts[0,0],0)]]
      closingstr=''.join([' L {x},{y}'.format(x=x, y=y) for x,y in [p1,p2]]+['z']) 
    pts[:,0],pts[:,1]=self.rescale(pts[:,0],pts[:,1])
    self.makePath(d=self.pathFromPts(pts, interp=interp)+closingstr, fill=fill,
        fill_opacity=fill_opacity, stroke=stroke,
        stroke_width=stroke_width,  
        marker_mid=self.getMarkerURL(marker=marker, markersize=markersize,
          stroke=stroke))
    plot_dict=dict(stroke=stroke, stroke_width=stroke_width, legend=legend, font_size=font_size,
        font_color=font_color, marker=marker, markersize=markersize)
    self.plots.append(plot_dict)
    if drawaxis:
      self.drawAxis()

  def drawAxis(self):
    sw=3
    self.makeLine(x1=sw/2, y2=3*sw, x2=sw/2, y1=self.height, stroke='black',
        stroke_width=sw, marker_end=self.getMarkerURL(marker='ft',
          markersize=sw, stroke='black', refX=-5))
    self.makeLine(x1=0, y1=self.height-sw/2, x2=self.width-3*sw, y2=self.height-sw/2,
        stroke='black', stroke_width=sw,
        marker_end=self.getMarkerURL(marker='ft', markersize=sw,
          stroke='black', refX=-5))
    xstep=10**(int(np.log10(self.xlim[1]-self.xlim[0])))
    xstart=(np.round(np.divide(self.xlim[0], xstep))+1)*xstep
    xstop=self.xlim[1]
    ystep=10**(int(np.log10(self.ylim[1]-self.ylim[0])))
    ystart=(np.round(np.divide(self.ylim[0], ystep))+1)*ystep
    strformat='%-3.1e'
    ystop=self.ylim[1]
    if xstep>1 and xstep<1000:
      strformat='%.0f'
    elif xstep>0.01 and xstep<=1:
      strformat='%.2f'
    if xstart+xstep>=xstop:
      xstep/=2.
      xstart-=xstep
    for x in np.arange(xstart, xstop, xstep):
      xpos=self.rescale(x=x)
      self.makeLine(x1=xpos, x2=xpos, y1=self.height, y2=self.height-8, stroke='black', stroke_width=sw)
      text=strformat%x
      bgdkwargs=dict(fill='grey', stroke='white', stroke_width=1, fill_opacity=0.8)
      self.makeText(text=text, x=xpos, y=self.height-20, font_size=15,
          stroke='white', fill='white', text_anchor='middle', dominant_baseline='central', bgdkwargs=bgdkwargs)
    strformat='%-3.1e'
    if ystep>1 and ystep<1000:
      strformat='%.0f'
    if ystart+ystep>=ystop:
      ystep/=2.
      ystart-=ystep
    for y in np.arange(ystart, ystop, ystep):
      ypos=self.rescale(y=y)
      self.makeLine(y1=ypos, y2=ypos, x1=0, x2=8, stroke='black', stroke_width=sw)
      ypos=self.rescale(y=y)
      text=strformat%y
      bgdkwargs=dict(fill='grey', stroke='white', stroke_width=1, fill_opacity=0.8)
      self.makeText(text=text, y=ypos, x=20, font_size=15, stroke='white',
          fill='white', text_anchor='middle', dominant_baseline='central', bgdkwargs=bgdkwargs)

  def setLims(self, xlim=None, ylim=None):
    if xlim is None:
      xlim=(0,float(self.width))
    elif np.isscalar(xlim):
      xlim=(0,float(xlim))
    if ylim is None:
      ylim=(0,float(self.height))
    elif np.isscalar(ylim):
      ylim=(0,float(ylim))
    self.xlim=xlim
    self.ylim=ylim

  def rescale(self, x=None, y=None):
    if x is not None:
      xx=(x-self.xlim[0])/(self.xlim[1]-self.xlim[0])*self.width
    if y is not None:
      yy=self.height-(y-self.ylim[0])/(self.ylim[1]-self.ylim[0])*self.height
    if x is None:
      return yy
    elif y is None:
      return xx
    else:
      return (xx,yy)

  def makeLegends(self):
    #!fixme: should have an argument about the placement ( automatic, upper-right, etc) 
    #        so far the legend is always at the upper-right corner
    margin=10
    linelength=20
    sep=5
    y=2*margin
    maxtextwidth=0
    height=0
    for p in self.plots:
      legend=p['legend']
      if legend is not None:
        font_size=p['font_size']
        height+=font_size+1
        textwidth=len(legend)*font_size*self.fontratio
        if textwidth>maxtextwidth:
          maxtextwidth=textwidth
    if maxtextwidth==0:
      return
    width=2*margin+maxtextwidth+linelength+sep
    left=self.width-width                        
    self.makeRectangle(x=left-margin, y=margin, width=width, height=height+margin,
        fill='grey', stroke='white', stroke_width=1, fill_opacity=0.8)
    for p in self.plots:
      legend=p['legend']
      if legend is not None:              
        stroke=p['stroke']
        stroke_width=p['stroke_width']
        font_size=p['font_size']
        font_color=p['font_color']
        marker=p['marker']
        markersize=p['markersize']
        self.makePath(d='M{left},{y} L{center},{y} L{right},{y}'.format(left=left, center=left+linelength/2.,
              right=left+linelength, y=y),
            fill=None, stroke=stroke, stroke_width=stroke_width,
            marker_mid=self.getMarkerURL(marker=marker, markersize=markersize, stroke=stroke))
        self.makeText(text=legend, x=left+linelength+sep,y=y,
              font_size=font_size, stroke=font_color, fill=font_color,
              dominant_baseline='central')
        y+=font_size+1

  def debug(self, mesg):
    self.debugMesgs.append(mesg)

  def makeDebugs(self, font_size=16): 
    if self.debugMesgs==[]:
      return
    bgdkwargs=dict(fill='blue', fill_opacity=0.95, stroke='white',
                  stroke_width=1)
    self.makeText(text=self.debugMesgs, x=self.width/2, y=self.height/2,
        font_size=font_size, fill='white',
        stroke='white', text_anchor='middle', dominant_baseline='central',
        bgdkwargs=bgdkwargs) 

           
  @classmethod
  def pathFromPts(cls,pts,interp='linear'):
    #!fixme: could add other interp modes (e.g nearest neighbor)
    if interp=='linear':
      d='\n'.join(['M%f,%f ' %(pts[0,0], pts[0,1])]+ 
            ['L%f,%f' %(pt2[0], pt2[1])
            for pt2 in pts[1:]])
    else: # this is cubic spline
      cpts1, cpts2=computeSplineControlPoints(pts)
      d='\n'.join( ['M%f,%f ' %(pts[0,0], pts[0,1])]+
          ['C%f,%f,%f,%f,%f,%f' %(cpt1[0],cpt1[1], cpt2[0],cpt2[1], pt2[0], pt2[1])
              for pt2, cpt1, cpt2 in zip(pts[1:], cpts1, cpts2)])
    return d

  @classmethod
  def pathFromFile(cls,filename):
    pts=np.loadtxt(filename)
    return cls.pathFromPts(pts)

  @classmethod
  def pathFromExpression(cls, func, xmin, xmax, xstep):
    pts=np.array([(x, func(x)) for x in np.arange(xmin, xmax, xstep)])
    return cls.pathFromPts(pts);

def test():
  scene = Scene("examples/test")
  p=SVGPlot(scene)
  p.makeRectangle(x=100,y=100,height=200,width=200,
         fill=(0,255,255),stroke=(0,0,0),
         stroke_width=1)
   
  p.makeLine(x1=200,y1=200,x2=200,y2=300,
        stroke=(0,0,0), stroke_width=1)
  p.makeLine(x1=200,y1=200,x2=300,y2=200,
        stroke=(0,0,0), stroke_width=1)
  p.makeLine(x1=200,y1=200,x2=100,y2=200,
        stroke=(0,0,0), stroke_width=1)
  p.makeLine(x1=200,y1=200,x2=200,y2=100,
        stroke=(0,0,0), stroke_width=1)
  p.makeCircle(cx=200, cy=200, r=30, fill=(0,0,255), stroke=(0,0,0), stroke_width=1)
  p.makeCircle(cx=200, cy=300, r=30, fill=(0,255,0), stroke=(0,0,0), stroke_width=1)
  p.makeCircle(cx=300, cy=200, r=30, fill=(255,0,0), stroke=(0,0,0), stroke_width=1)
  p.makeCircle(cx=100, cy=200, r=30, fill=(255,255,0), stroke=(0,0,0), stroke_width=1)
  p.makeCircle(cx=200, cy=100, r=30, fill=(255,0,255), stroke=(0,0,0), stroke_width=1)
   
  p.makeText(x=50,y=50,text="Testing SVG",font_size=24,fill=(0,0,0))
   
  p.makeImage(href='examples/portrait.jpg', x=0, y=10, width=100, height=100)
  scene.write_svg()
  scene.display()
  return

  

def test2():
  width,height=309,517
  scene = Scene("examples/test2",width=width, height=height)
  p=SVGPlot(scene)
  p.makeText(x=50,y=50,text="Testing SVG",font_size=24,fill=(0,0,0))
  p.makeImage(href='../examples/shannon.png', x=0, y=0, width=width, height=height)
  pts=np.loadtxt('examples/shannonMSE.data')
  pts=pts[1::2]
  pts2=np.array(pts)
  pts2[:,1]+=10
  pts2[:,0]+=10
  p.plot(pts, interp='cubicspline', legend='RMSE', marker='fs', markersize=3)
  p.plot(pts2, legend='upper', stroke='green', marker='ft', drawaxis=True)
  p.makeLegends()
  scene.write_svg()
  scene.display()
  #scene.convert()
  return 

if __name__ == "__main__": test2()
