import os
import sys
import shutil
import numpy as np
import PIL.Image
from SVGPlot import Scene, SVGPlot

class Plotter:
    def __init__(self, work_dir='./', rx=2, names=['shannon']):
        self.work_dir=work_dir
        self.rx=rx
        self.names=names

    def makePlots(self):
        size=PIL.Image.open(self.work_dir+'gt.png').size
        width=size[0]
        height=size[1]
        # plot the RMSE
        base_dir = os.path.dirname(os.path.abspath(__file__))
        shutil.copy(base_dir+'/SVGPlot/SVGPan.js', self.work_dir)
        prevdir=os.getcwd()
        os.chdir(self.work_dir)
        self.makeRMSEPlots(width=width, height=height)
        self.makeRMSEPlots(width=width, height=height, envelop=True)
        self.makeSpectrumPlot(width=width, height=height)
        self.makeCumSpectrumPlot(width=width, height=height)
        os.chdir(prevdir)

    def makeSpectrumPlot(self, width, height):
        scene=Scene('Spectrum', width=width, height=height, pathToJS='SVGPan.js')
        p=SVGPlot(scene)
        pts=np.loadtxt('spectrum.data')
        pts[:,1]=10*np.log10(pts[:,1])# don't use 20*log since it is a power spectrum already
        fillkwargs=dict(fill='blue', fill_opacity=0.5)
        p.plot(pts[1:], stroke='blue', interp='linear', xlim=pts[-1,0],
            ylim=pts[:,1].max(), legend='Spectrum (dB)',
            drawaxis=True, **fillkwargs)
        r=4
        k1,k2=2, np.argmax(pts[-r:,1])-r
        for k in [k1,k2]:
           xm,ym=pts[k,0],pts[k,1]
           if k==k1:
               x2,y2=p.rescale(x=xm)+50, p.rescale(y=ym)
           else:
               x2,y2=p.rescale(x=xm)-25, p.rescale(y=ym)-50
           p.makeLine(x1=p.rescale(x=xm), x2=x2,
               y1=p.rescale(y=ym), y2=y2,
                 stroke='blue',stroke_width=1, stroke_dasharray='9, 5')
           font_size=16
           text='%.1fdB'%ym
           bgdkwargs=dict(fill='gray', fill_opacity=0.8, stroke='blue',
               stroke_width=1, stroke_dasharray='9, 5')
           p.makeText(text=text, x=x2, y=y2,
               font_size=font_size, fill='white',
               stroke='white', text_anchor='middle', dominant_baseline='central',
               bgdkwargs=bgdkwargs)
        p.makeLegends()
        p.makeDebugs()
        scene.write_svg()

    def makeCumSpectrumPlot(self, width, height):
        scene=Scene('CumSpectrum', width=width, height=height, pathToJS='SVGPan.js')
        p=SVGPlot(scene)
        pts=np.loadtxt('spectrum.data')
        spectrumVals=np.array(pts[:,1])
        pts[:,1]=np.cumsum(spectrumVals)
        fillkwargs=dict(fill='blue', fill_opacity=0.5)
        p.plot(pts[1:], stroke='blue', interp='linear', xlim=pts[-1,0],
            ylim=pts[:,1].max(), legend='Cumulative spectrum',
            drawaxis=True, **fillkwargs)
        # compute the tangent at Nyquist rate
        r=10
        fEnd=pts[-1,1]
        fPrime=np.mean(spectrumVals[-r:])
        wNyq=pts[-1,0]
        fStart=fEnd-(len(spectrumVals)-1)*fPrime
        p.makeLine(x1=p.rescale(0), y1=p.rescale(y=fStart),
                x2=p.rescale(x=wNyq), y2=p.rescale(y=fEnd),
                stroke='green', stroke_width=3, stroke_dasharray='9, 5')
        # copmute a pvalue
        ratio=0.99
        k=np.argmin(pts[:,1]<=ratio*pts[-1,1])
        xk=pts[k,0]
        yk=pts[k,1]
        p.makeLine(x1=p.rescale(x=xk), y1=p.rescale(y=0),
                   x2=p.rescale(x=xk), y2=p.rescale(y=yk),
                   stroke='red',stroke_width=3, stroke_dasharray='9, 5')
        font_size=16
        text='{ratio}% pvalue'.format(ratio=int(100*ratio))
        bgdkwargs=dict(fill='gray', fill_opacity=0.8, stroke='red',
            stroke_width=3, stroke_dasharray='9, 5')
        p.makeText(text=text, x=p.rescale(x=xk), y=p.rescale(y=0)-100,
            font_size=font_size, fill='white',
            stroke='white', text_anchor='middle', dominant_baseline='central',
            bgdkwargs=bgdkwargs)
        p.makeLegends()
        p.makeDebugs()
        scene.write_svg()

    def makeRMSEPlots(self, width, height, envelop=False):
        rx = self.rx
        pts=np.loadtxt('shannonUB.data')
        xmax=np.max(pts[:,0])
        ymax=np.max(pts[:,1])
        for name in self.names:
          outname=name+'RMSE'
          fillkwargs={}
          if envelop:
            outname=name+'Envelop'
          scene=Scene(outname, width=width,
              height=height,pathToJS='SVGPan.js')
          p=SVGPlot(scene)
          p.makeImage(href=name+'.png', x=0, y=0, width=width, height=height)
          if name in ['shannon', 'cshannon', 'dft']:
            pts2=np.loadtxt(name+'UB.data')
            if envelop:
              k0=np.argmin(np.abs(pts2[:,0]-1))
              pts2=pts2[k0::-rx]
            fillkwargs=dict(fill='green', fill_opacity=0.5)
            p.plot(pts2, stroke='green', interp='cubicspline', xlim=(0.5,
              2*xmax), ylim=ymax, legend='upper bound', **fillkwargs)
          pts=np.loadtxt(name+'MSE.data')
          pts[:,1]=np.sqrt(pts[:,1])
          pts[:,0]=pts[:,0]/rx+0.5
          if envelop:
            k0=np.argmin(np.abs(pts[:,0]-1))
            pts=pts[k0::rx]
          fillkwargs=dict(fill='red', fill_opacity=0.5)
          p.plot(pts, stroke='red', interp='cubicspline', xlim=(0.5, 2*xmax),
              ylim=ymax, legend='empirical estimate', drawaxis=True, **fillkwargs)
          # show some values of the RMSE
          if name != 'gt':
            k1,k2,k3=np.argmax(pts[:,1]), np.argmax(pts[5:,1])+5, np.argmax(pts[-rx:,1])-rx
            if envelop: k3=len(pts)-1
            for k in [k1,k2,k3]:
              xm,ym=pts[k,0],pts[k,1]
              p.makeLine(x1=p.rescale(x=xm), x2=p.rescale(x=xm)+50,
                  y1=p.rescale(y=ym), y2=p.rescale(y=ym)-50,
                    stroke='red',stroke_width=1, stroke_dasharray='9, 5')
              font_size=16
              text='rmse=%.1f'%ym
              bgdkwargs=dict(fill='gray', fill_opacity=0.8, stroke='red',
                  stroke_width=1, stroke_dasharray='9, 5')
              p.makeText(text=text, x=p.rescale(x=xm)+50, y=p.rescale(y=ym)-50,
                  font_size=font_size, fill='white',
                  stroke='white', text_anchor='middle', dominant_baseline='central',
                  bgdkwargs=bgdkwargs)
          p.makeLegends()
          p.makeDebugs()
          scene.write_svg()
          #scene.convert()

if __name__ == '__main__':
    n=len(sys.argv)
    if n>1:
        work_dir=sys.argv[1]
    else:
        work_dir='./'
    if n>2:
        rx=int(sys.argv[2])
    else:
        rx=2
    if n>3:
        names=sys.argv[3:]
    else:
        names=['shannon']
    ## print('{wd} {rx} {names}'.format(wd=work_dir, rx=rx, names=names))
    p=Plotter(work_dir, rx, names)
    p.makePlots()
